

import os

INTERESTING = 0x6dc 
INTERESTING_ADDRESSES = [ 0x706 ]
INTERESTING_STORE_VALUES = []
DO_NULL_OPEN = True
CHECK_REFS = False
END_OF_WORLD = 0x200
SPARE_SPACE = 0x20

def Scan_File(bug_out, occlude_in, asm_out, py_out):
    expect_data = dict()
    occlude_out = set()
    region = dict()
    limit = 0
    case_number = 0

    def Comment(text):
        asm_out.write("/* %s */\n" % text)
        py_out.write("# %s\n" % text)

    def Whats_Open():
        if (case_number >= INTERESTING):
            for (ref, (begin, end, open_data)) in region.iteritems():
                Comment("O %(expect_ref)d (index) %(base)x to "
                    "%(end1)x, copy location %(copy_loc)x" % open_data)
                Comment("O %(expect_ref)d (void*) %(address)x to "
                    "%(end2)x" % open_data)

    old_case_number = 0

    for line in file(bug_out):
        fields = line.split()
        if len(fields) < 4:
            continue

        if (case_number != old_case_number):
            py_out.write(" print 'Case %d (0x%x)'\n" % (case_number, case_number))
            old_case_number = case_number

        line = line.rstrip()
        command = fields[ 1 ]
        m1command = None

        if command in ("word", "block"):
            m1command = fields[ 0 ]

        if (command == "OPEN") or (m1command == "open"):
            print '%x: %s' % (case_number, line)
            if m1command:
                address = (int(fields[ 4 ], 16) >> 2) & 0x0fff
                size = int(fields[ 6 ], 16) >> 2
                copy_loc = int(fields[ 8 ], 16)
                expect_ref = int(fields[ 11 ], 16)
                if fields[ 9 ] == "RO":
                    read_only = 1
                elif fields[ 9 ] == "RW":
                    read_only = 0
                else:
                    assert False
            else:
                address = int(fields[ 2 ], 16)
                size = int(fields[ 4 ], 16)
                copy_loc = int(fields[ 6 ], 16)
                expect_ref = int(fields[ 10 ], 16)
                read_only = 0

            assert copy_loc + size + SPARE_SPACE <= END_OF_WORLD
            assert 0 <= copy_loc
            assert 0 < size 

            interesting = (not case_number in occlude_in)
            Whats_Open()
            open_data = {
                "case_number" : case_number,
                "control_word" : (read_only << 31) | (size << 16) | copy_loc,
                "expect_ref" : expect_ref,
                "address" : address * 4,
                "base" : address,
                "end1" : address + size,
                "end2" : (address + size) * 4,
                "size" : size,
                "copy_loc" : copy_loc,
                "read_only" : read_only,
                "interesting" : interesting,
            }
            if interesting:
                Comment("OPEN - case number %(case_number)d "
                    "/ 0x%(case_number)x" % open_data)
                Comment("O %(expect_ref)d (index) %(base)x to "
                    "%(end1)x, copy location %(copy_loc)x" % open_data)
                Comment("O %(expect_ref)d (void*) %(address)x "
                    "to %(end2)x" % open_data)
                py_out.write(" (open_time, ref) = yield st.Open("
                    "0x%(base)x, 0x%(size)x, 0x%(copy_loc)x, "
                    "read_only = %(read_only)u)\n" % open_data)
                py_out.write(" ref_table[%(expect_ref)u] = "
                    "ref\n" % open_data)
                asm_out.write("""
open_case_%(case_number)x:
    addi    r8, r0, 0x%(case_number)x
    addi    r4, r0, 0x%(control_word)x
    swi     r4, r2, 0           /* Write control */
    lwi     r5, r3, 0x%(address)x  /* OPEN */
    lwi     r7, r2, 0x10        /* Read error code */
    andi    r9, r7, 0xff
    xori    r9, r9, 0x09
    bnei    r9, failure
    sbi     r5, r16, 0x%(expect_ref)x
""" % open_data)
            if CHECK_REFS:
                asm_out.write("""
    xori    r6, r5, 0x%(expect_ref)x
    bnei    r6, failure
""" % open_data)

            limit = max(address + size, limit)
            region[ expect_ref ] = (address, address + size, open_data)
            case_number += 1

                    

        elif (command == "CLOSE") or (m1command == "close"):
            print '%x: %s' % (case_number, line)
            if m1command:
                ref = int(fields[ 4 ], 16)
            else:
                ref = int(fields[ 3 ], 16)
            Whats_Open()
            (address, end, open_data) = region[ ref ]

            if case_number < INTERESTING:
                occlude_out.add(open_data[ "case_number" ])

            case_number += 1
            open_data[ "case_number" ] = case_number
            assert open_data[ "expect_ref" ] == ref
            open_data[ "ref" ] = ref
            interesting = open_data[ "interesting" ]
            if interesting:
                Comment("CLOSE - case number %(case_number)d "
                        "/ 0x%(case_number)x" % open_data)
                Comment("C %(ref)d (index) %(base)x to %(end1)x, "
                        "copy location %(copy_loc)x" % open_data)
                Comment("C %(ref)d (void*) %(address)x to "
                        "%(end2)x" % open_data)
                py_out.write(" yield st.Close(ref_table"
                        "[%(expect_ref)u])\n" % open_data)
                asm_out.write("""
close_case_%(case_number)x:
    addi    r8, r0, 0x%(case_number)x""" % open_data)

                if CHECK_REFS:
                    asm_out.write("""
    addi    r5, r0, 0x%(ref)x""" % open_data)
                else:
                    asm_out.write("""
    lbui    r5, r16, 0x%(expect_ref)x""" % open_data)

                asm_out.write("""
    swi     r0, r2, 0           /* Write control */
    swi     r5, r3, 0           /* CLOSE */
    lwi     r7, r2, 0x10        /* Read error code */
    andi    r9, r7, 0xff
    xori    r9, r9, 0xc9
    bnei    r9, failure
""" % open_data)
            del region[ ref ]

        elif m1command == "store":
            address = (int(fields[ 3 ], 16) >> 2) & 0x0fff
            value = int(fields[ 5 ], 16)
            limit = max(address + 4, limit)
            if (((value in INTERESTING_STORE_VALUES)
            or (address in INTERESTING_ADDRESSES))
            and (case_number >= INTERESTING)):
                print '** Interesting value 0x%x written to 0x%x' % (value,
                        address)
                expect_data[ address ] = value
                store_data = {
                    "value" : value,
                    "word" : address,
                    "address" : address * 4,
                }
                Comment("STORE index %(word)x" % store_data)
                py_out.write(" yield st.Store(0x%(word)x, "
                        "0x%(value)x)\n" % store_data)
                asm_out.write("""
    addi    r4, r0, 0x%(value)x
    swi     r4, r3, 0x%(address)x
""" % store_data)

    Whats_Open()
    
    return (expect_data, occlude_out, region, limit, case_number)

def Main(bug_out):

    py_out = file("../x.py", "wt")
    py_out.write("""

from twisted.internet import defer

@defer.inlineCallbacks
def OcTest(st):
 ref_table = dict()
""")

    asm_out = file("x.s", "wt")
    asm_out.write("""
    
.global octest
.global ocresults
.bss
lmb:
    .space  0x4000
ocresults:
    .space  0x100
ref_table:
    .space  0x40
.text
octest:
    addi    r1, r1, -0x100
/* Save and zero registers */
""")
    for i in xrange(2, 32, 1):
        asm_out.write("""
    swi     r%d, r1, 0x%x""" % (i, i * 4))

    for i in xrange(2, 32, 1):
        asm_out.write("""
    add     r%d, r0, r0""" % (i))

    asm_out.write("""
/* Load invariants */
    addi    r2, r0, 0xa0000000
    addi    r3, r0, lmb
    addi    r16, r0, ref_table

/* Zero memory space */
    addi    r4, r0, 0x4000
0:
    addi    r4, r4, -4
    sw      r4, r3, r4
    bnei    r4, 0b

""")
# r2 = control address
# r3 = base of LMB
# r4 = control word (temporary)
# r5 = reference (from OPEN)
# r6 = comparison result (r5 versus expected reference)
# r7 = error code
# r8 = case number
# r9 = comparison result (r7 versus expected code)
# r10 = access address
# r11 = value from LOAD
# r12 = status
# r13 = temporary copy of "load time"
# r14 = expected value from LOAD
# r15 = comparison result (load expected)
# r16 = base of reference table
    if DO_NULL_OPEN:
        read_only = 0
        size = 4 # spare_space
        copy_loc = END_OF_WORLD - size 
        asm_out.write("""
/* Null "OPEN" */
dontcare:
    addi    r8, r0, -1
    addi    r4, r0, 0x%(control_word)x
    swi     r4, r2, 0           /* Write control */
    lwi     r5, r0, dontcare    /* Address */
    xori    r6, r5, 0xf
    bnei    r6, failure
""" % { 
        "control_word" : (read_only << 31) | (size << 16) | copy_loc,
    })
        py_out.write(" # null open\n")
        py_out.write(" yield st.Open(0x10000, 0x%x, 0x%x)\n" % (size, copy_loc))

    py_out.write("\n\n")

    occlude = set()
    (expect_data, occlude, region, limit, case_number) = Scan_File(
                bug_out, occlude, os.tmpfile(), os.tmpfile())
    (expect_data, occlude, region, limit, case_number) = Scan_File(
                bug_out, occlude, asm_out, py_out)
           
    py_out.write("\n\n # tests\n\n")

    print 'checking to limit 0x%x' % limit
    asm_out.write("""

___begin_tests:
/* Still OPEN: */ """)

    should_be_open = set()
    todo = [ 0 ]

    for (ref, (begin, end, open_data)) in region.iteritems():
        for address in xrange(begin, end, 1):
            should_be_open.add(address)

        todo.append(begin - 1)
        todo.append(begin)
        todo.append(end - 1)
        todo.append(end)
        todo.append(begin & 0x3ff)
        todo.append(begin & 0x1ff)

    asm_out.write("\n")

    todo.extend(range(0, limit, 1))
    done = set()
    error = 0
    for address in todo:
        if address in done:
            continue
        if ((address < 0) or (address >= limit)):
            continue

        done.add(address)

        if address in should_be_open:
            branch_type = "bnei"
        else:
            branch_type = "beqi"

        value = expect_data.get(address, address * 4)
        value2 = expect_data.get(address, 0xffff0000 | address)

        load_data = {
                "case_number" : case_number,
                "address" : address * 4,
                "word" : address,
                "branch_type" : branch_type,
                "value" : value,
                "value2" : value2,
                }
        asm_out.write("""

/* Test 0x%(address)x (word 0x%(word)x) - case number %(case_number)d / 0x%(case_number)x */
    lwi     r24, r2, 0x04       /* Global counter - before */
    lwi     r26, r2, 0x08       /* Match counter - before */
    lwi     r28, r2, 0x0c       /* Miss counter - before */
    addi    r8, r0, 0x%(case_number)x
    lwi     r11, r3, 0x%(address)x  /* LOAD */
    lwi     r12, r2, 0          /* Read status */
    lwi     r7, r2, 0x10        /* Read error code */
    lwi     r25, r2, 0x04       /* Global counter - after */
    lwi     r27, r2, 0x08       /* Match counter - after */
    lwi     r29, r2, 0x0c       /* Miss counter - after */
    andi    r13, r12, 0x7fff    /* Mask asm_out "load time" bits */
    bsrli   r13, r13, 2         /* load time divided by 4 */
    %(branch_type)s r13, failure
    addi    r14, r0, 0x%(value)x   /* check value stored in memory */
    xor     r15, r14, r11
    bnei    r15, failure

""" % load_data)

        py_out.write("""
 # Test 0x%(address)x (word 0x%(word)x) - case number %(case_number)d / 0x%(case_number)x 
 (load_time, value) = yield st.Load(0x%(word)x)
 assert value == 0x%(value2)x, hex(value)
""" % load_data) 
        if address in should_be_open:
            py_out.write(" assert load_time < 4 # OPEN\n")
        else:
            py_out.write(" assert load_time >= 4 # CLOSEd\n")
        case_number += 1


    py_out.write(" print 'Success: 1337 1337'\n\n\n\n")
    asm_out.write("""
/* Success */
    addi    r8, r0, 0x13371337
    or      r30, r0, r8
    or      r31, r0, r8

failure:
/* Copy asm_out results, restore registers*/
""" % { "case_number" : case_number,
    })

    for i in xrange(0, 32, 1):
        asm_out.write("""
    swi     r%d, r0, ocresults + 0x%x""" % (i, i * 4))

    for i in xrange(2, 32, 1):
        asm_out.write("""
    lwi     r%d, r1, 0x%x""" % (i, i * 4))

    asm_out.write("""
    rtsd    r15, 0x8
    addi    r1, r1, 0x100
""")
    print 'success case is %d' % case_number



Main("cmd2.txt")

