/*
 * Scratchpad MMU - SMMU hardware driver for JPEG case study
 * Copyright (C) Jack Whitham 2009
 * http://www.jwhitham.org.uk/c/smmu.html
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation
 * (version 2.1 of the License only).
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
#ifndef DRVIF_H
#define DRVIF_H

#include "../driver/hwdriver.h"
#include <assert.h>

#define MAX_AR_ID           10

#define TIME_DCT            0
#define TIME_CC             1
#define TIME_DCT1           2
#define TIME_DEC            3

#define ACT_DCT             1
#define ACT_DCTOUT          2
#define ACT_YCC             4
#define ACT_DEC             8
#define ACT_DEC1            16

typedef struct Activity_Record_struct {
        unsigned        global_counter ;
        unsigned        match_count ;
        unsigned        miss_count ;
        unsigned        call_count ;
} Activity_Record ;


typedef struct Stack_struct {
        unsigned char ptr ;
        unsigned char ref [ MAX_NUM_ENTRIES ] ;
        unsigned short loc ;
        unsigned short size [ MAX_NUM_ENTRIES ] ;
} Stack ;

extern Stack _open_stack ;
extern Activity_Record _ar_start [ MAX_AR_ID ] ;
extern Activity_Record _ar_id [ MAX_AR_ID ] ;
extern unsigned activate_smmu ;
extern unsigned read_only_remaining;





static inline unsigned HW_OPEN_WITH_COMMENT ( void * address , 
            int offset , unsigned size , unsigned copy_loc ,
            const char * comment , int ro )
{
    /* CONTROL_OFFSET = (unsigned) offset; */
    /* CONTROL_COMMENT = (unsigned) comment; */
    address = (void *) ((unsigned) address + (unsigned) offset);
    /* printf ( "\r\nOPEN(%p, 0x%x, 0x%x);\r\n", address, size, copy_loc); */
    if ( ro && ( read_only_remaining != 0 ))
    {
        read_only_remaining --;
        return HW_OPEN_RO(address, size, copy_loc);
    } else {
        return HW_OPEN(address, size, copy_loc);
    }
}

static inline int HW_OPEN_STACK ( void * base_pointer , 
            int offset , unsigned size ,
            const char * comment , int ro )
{
    unsigned size1 = ( size + 3 ) >> 2 ;
    unsigned ref;

    assert ( _open_stack . ptr < NUM_ENTRIES ) ;
    assert (( _open_stack . loc + size1 ) <= SP_SIZE ) ;
    _open_stack . size [ _open_stack . ptr ] = _open_stack . loc ;
    ref = HW_OPEN_WITH_COMMENT ( base_pointer , offset , size , 
                    _open_stack . loc , comment , ro ) ;
    if (( ref == 0 )
    || ( ref >= NUM_ENTRIES ))
    {
        return 0;
    }
    _open_stack . ref [ _open_stack . ptr ] = ref;
    _open_stack . ptr ++ ;
    _open_stack . loc += size1 ;
    return 1;
}

static inline void HW_CLOSE_STACK ( void )
{
    unsigned ref;
    assert ( _open_stack . ptr > 0 ) ;
    _open_stack . ptr -- ;
    _open_stack . loc = _open_stack . size [ _open_stack . ptr ] ;
    ref = _open_stack . ref [ _open_stack . ptr ] ;
    assert ( ref != 0 ) ;
    assert ( ref < NUM_ENTRIES ) ;
    HW_CLOSE ( ref );
    assert ( _open_stack . loc <= SP_SIZE ) ;
    if ( ref < FIRST_RW_ENTRY )
    {
        read_only_remaining ++;
    }
}

static inline void Capture_Activity_Record ( 
        Activity_Record * out , Activity_Record * sub )
{
    out -> global_counter += 
                    CONTROL -> global_counter - sub -> global_counter ;
    out -> match_count += CONTROL -> match_count - sub -> match_count ;
    out -> miss_count += CONTROL -> miss_count - sub -> miss_count ;
}

static inline void Start_Timing ( unsigned id )
{
    Capture_Activity_Record ( & _ar_start [ id ] , & _ar_start [ id ] ) ;
}

static inline void Stop_Timing ( unsigned id )
{
    assert ( id < MAX_AR_ID ) ;
    Capture_Activity_Record ( & _ar_id [ id ] , & _ar_start [ id ] ) ;
    _ar_id [ id ] . call_count ++ ;
}

void Relocate ( void ) ;
void Zero_Spad ( void ) ;


#endif

